/*
 *  MP3_14 Handy Controller
 */
#include <18F26J50.h>

#device ansi
#device const=read_only
#device PASS_STRINGS=IN_RAM
#opt compress

//#fuses HSPLL,PLL1,NOCPUDIV,NOFCMEN,NOIESO,NOWDT,STVREN,NOXINST,NODEBUG
//#fuses HS,NOCPUDIV,NOFCMEN,NOIESO,NOWDT,STVREN,NOXINST,NODEBUG
//#fuses INTRC_IO,NOCPUDIV,NOFCMEN,NOIESO,NOWDT,STVREN,NOXINST,NODEBUG
//#use delay (clock=8000000)

//#fuses INTRC_PLL_IO,PLL2,CPUDIV3,NOFCMEN,NOIESO,NOWDT,STVREN,NOXINST,NODEBUG
#fuses INTRC_IO,PLL2,CPUDIV3,NOFCMEN,NOIESO,NOWDT,STVREN,NOXINST,NODEBUG
#use delay (clock=16000000)

#bit USBEN   = 0x0F65.3       // 
#bit PEIE    = 0x0FF2.6 

// USART2 : ESP-WROOM-02操作
#pin_select TX2=PIN_A1
#pin_select RX2=PIN_A0
#byte TXSTA2    = 0x0FA8
#byte RCSTA2    = 0x0F9C
#byte BAUDCON2  = 0x0F7C
#byte SPBRGH2   = 0x0F7D 
#byte SPBRG2    = 0x0FAB 
#bit RC2IF      = 0x0FA4.5
#bit TX2IF      = 0x0FA4.4
#byte TXREG2    = 0x0FA9
#byte RCREG2    = 0x0FAA
#bit TRISA0     = 0x0F92.0
#bit TRISA1     = 0x0F92.1
#bit RC2IE      = 0x0FA3.5

// OLED操作
#use I2C(MASTER,SCL=PIN_A2,SDA=PIN_A3,FORCE_SW,FAST,STREAM=OLED)
//#use I2C(MASTER,SCL=PIN_A2,SDA=PIN_A3,FORCE_SW,SLOW,STREAM=OLED)

// デバッグ用
#use rs232(STREAM=DEBUG,BAUD=9600,XMIT=PIN_C6,RCV=PIN_C4,FORCE_SW,PARITY=N,BITS=8,STOP=1,ERRORS)

#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include "integer.h"
#include "diskio.h"
#include "ffconf.h"
#include "ff.h"
#include "Mp3_14_hc.h"

// ディレクトリ構成ファイル
typedef struct _DirList {
    uint16_t    recno_subdir_start;     // サブディレクトリの先頭レコード番号
    uint8_t     num_subdir;             // サブディレクトリ数
    uint16_t    recno_parentdir;        // 親ディレクトリのレコード番号
    uint16_t    recno_grandparentdir;   // 親々ディレクトリのレコード番号
    uint8_t     num_sndfile;            // MP3ファイル数
    uint16_t    recno_mp3list;          // MP3ファイルリスト先頭レコード番号
} DIRLIST;

// Timer
#define CHATTERING_ADJUST           15
#define TIMER0_INIT                 55536

// POWER
#define PIN_POWER_ON_MAIN       PIN_C1      //  (output)
#define PIN_DETECT_POWER_OFF    PIN_C2      // detect power off (input)
#define PIN_LED                 PIN_A5

// Switch PIN
#define PIN_SWDOWN              PIN_B7
#define PIN_SWUP                PIN_B6
#define PIN_SWSELECT            PIN_B3
#define PIN_SWRETURN            PIN_B2
#define PIN_SWPLAY              PIN_B1
#define PIN_SWRANDOM            PIN_B0

#define sw_down()               input(PIN_SWDOWN)
#define sw_up()                 input(PIN_SWUP)
#define sw_select()             input(PIN_SWSELECT)
#define sw_return()             input(PIN_SWRETURN)
#define sw_play()               input(PIN_SWPLAY)
#define sw_random()             input(PIN_SWRANDOM)
#define sw_poff()               input(PIN_DETECT_POWER_OFF)

// Event
#define SWDOWN_PRESSED          0x80
#define SWUP_PRESSED            0x81
#define SWSELECT_PRESSED        0x82
#define SWRETURN_PRESSED        0x83
#define SWPLAY_PRESSED          0x84
#define SWRANDOM_PRESSED        0x85
#define SWPOFF_PRESSED          0x86
#define NOWPLAYING_DETECTED     0x87
#define WIFI_ERROR_DETECTED     0x88

// Action
#define PLAY_OTHER_DIR          0x93
#define POWEROFF                0x94
#define WIFI_RESTART            0x95

// Size
#define FILERWBUFSIZE           256
#define WIFISRBUFSIZE           200    //USARTの受信データのバッファサイズ(Byte)
#define LCD_WIDTH               128
#define LCD_PAGES               8
#define NUM_OLED_LINES          4           // 16ドットフォントで表示できる行数
#define TIMESTAMP_LEN           14

// MP3_14 コマンド
#define MP3_14_CMD_PLAYDIR          0x01
#define MP3_14_CMD_PLAYRANDOM       0x03
#define MP3_14_CMD_VOLUP            0x04
#define MP3_14_CMD_VOLDOWN          0x05
#define MP3_14_CMD_PLAYSTOP         0x06
#define MP3_14_CMD_PLAYSKIP         0x07
#define MP3_14_CMD_ACTIVE           0x08
#define MP3_14_CMD_NOWPLAYING       0x09
#define MP3_14_CMD_WAITING          0x0A
#define MP3_14_CMD_NEXT             0x0B
#define MP3_14_CMD_SHUTDOWN         0x0D
#define MP3_14_CMD_DATATIMESTAMP    0x10
#define MP3_14_CMDLENGTH            16           // ソケットを通じて送受信する１単位のデータサイズ(BYTE)
#define MP3_14_VOLUME_STEP          2            // 音量の増減単位

// WIFI
#define WIFI_OK         0
#define WIFI_ERROR      1
#define WIFI_TIMEOUT    2
#define WIFI_FULL       3

// OLED
#define OLED_ADDRESS        0x3C
#define OLED_CB_COMMAND     0x00
#define OLED_CB_COMMAND_CO  (0x00 | 0x80)
#define OLED_CB_DATA        0x40
#define OLED_CB_DATA_CO     (0x40 | 0x80)

// ルーター／サーバー設定
#define ROUTERFILE              "/ROUTER.TXT"

// ディレクトリ構成ファイル DIRLISTFILE parameters
#define DIRLISTFILE             "/DIRLST14.DAT"
#define RECLEN_DIRLIST           768UL
#define RECLEN_DIRLIST_FULL      (RECLEN_DIRLIST + 256)    // 全レコード長
#define DL_RECNO_SUBDIR_START    691
#define DL_NUM_SUBDIR            693
#define DL_RECNO_PARENTDIR       694
#define DL_RECNO_GRANDPARENTDIR  696
#define DL_NUM_SNDFILE           698
#define DL_RECNO_MP3LIST         699
#define RECLEN_IMAGEDATA         256

// 音源ファイルリスト
#define	SNDFILELISTFILE         "/SNDLST14.DAT"			// MP3ファイルリスト(Version=14)
#define	RECLEN_MP3LIST			(512UL)
#define	RECLEN_MP3LIST_FULL		(RECLEN_MP3LIST + 256)  // 全レコード長
#define ML_TITLENAME_OFFSET		0
#define ML_TITLENAME_LEN		128
#define ML_FILENAME_OFFSET		128
#define ML_FILENAME_LEN			(MP3FILE_LEN + 1)
#define ML_FILLER_14			191
#define ML_IMAGEDATA_OFFSET		512UL					// LCD表示のイメージデータ
#define ML_IMAGEDATA_LEN		256UL

// フォントファイル（半角）
#define FONTFILE                "/MYRCH16B.FNT"
#define FONT_WIDTH              8
#define FONT_HEIGHT             16
#define FONT_HEIGHT_BYTE        2
#define FONT_BYTES_CHAR         (FONT_HEIGHT_BYTE * FONT_WIDTH)

// メッセージ
#define MSG_SETTING_ERROR           "Setting File Err"
#define MSG_CONNECTED_ROUTER        "Router Connected"
#define MSG_NO_ROUTER               "No Router       "
#define MSG_MP3_14_EXISTS           "MP3_14 Connected"
#define MSG_MP3_14_NOT_EXISTS       "No MP3_14       "
#define MSG_WIFI_ERROR              "WIFI ERROR !    "
#define MSG_PLAYDIR_ERROR           "PLAY DIR ERROR  "
#define MSG_DATAFILE_IS_OLD         "Datafile is OLD "

#define DIGIT_CHAR                  "0123456789: /"

// 
static FATFS        Fatfs;
static BOOLEAN      EVENT_SWDOWN;
static BOOLEAN      EVENT_SWUP;
static BOOLEAN      EVENT_SWSELECT;
static BOOLEAN      EVENT_SWRETURN;
static BOOLEAN      EVENT_SWPLAY;
static BOOLEAN      EVENT_SWRANDOM;
static BOOLEAN      EVENT_SWPOFF;
static uint16_t     CHATTERING_TMR;

unsigned int16      WaitMsCycle;
volatile int16      WifiTimeoutCounter;
static BOOLEAN      ConnectedMP3_14;

/*
 *  IPD取得処理前、"+"読み込み後、"+IPD:"読み込み後、"+IPD:データ長,"読み込み後、データ読み込み後
 */
enum ipdstatus_t    {IPD_INIT, IPD_RECVD_PLUS, IPD_RECVD_IPD, IPD_RECVD_LEN, IPD_RECVD_DATA};

volatile char WifiSrBuf[WIFISRBUFSIZE];     // WIFIバッファ
volatile int WifiSrIdx = 0;
volatile int WifiSrIdx_IPD;
volatile int RecvedFlag;
volatile int RecvedBytes;
volatile int PromptFlag;
volatile ipdstatus_t IPDStatus = IPD_INIT;
int IPDDataLen;
volatile int16   SendActiveCnt;
volatile BOOLEAN DatatimestampRecv;

uint8_t    DigitImageData[8*2*13];       // DIGIT_CHARのイメージデータ(1文字の幅８ドット、高さ2バイト、13文字)

void InitApp(void);
void wait_ms(unsigned int16 ms);
int8 ReadRouterFile(FIL *Fil, uint8_t *buf, int16 bufsize, char *ssid, char *passwd, char *mp3_14_ip, char *mp3_14_port);
int wifi_connect_router(char *ssid, char *passwd);
int wifi_connect_server(char *remorte_ip, char * remorte_port);
int wifi_send_command(char* cmd, int bytes);
int wifi_restart();
void wifi_putstr(char *str);
int wifi_waitOk(long timeout);
int wifi_waitPrompt(long timeout);
void wifi_readclear();
void uart_test();
void usartCharSend(char c);
void serial_rda_interrupt();
void InitOled();
void OledTest();
BOOLEAN GetMasterFiles(FIL *Fil, uint8_t *FileRwBuf);
int8 SelectDirectory(FIL *Fil, FIL *Fil2, uint8_t *FileRwBuf, uint8_t *FileRwBuf2);
int8 SelectMp3InDirectory(FIL *Fil, FIL *Fil2, uint8_t *FileRwBuf, uint8_t *FileRwBuf2, uint16_t recno_first, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno);
void SelectShutdown(FIL *Fil, uint8_t *FileRwBuf);
FRESULT f_read_DIRLISTFILE (FIL* fp,	void* buff, DIRLIST* diritem);
int Mp314_PlayDir(uint16_t dirno, uint8_t trackno, uint16_t start_time);
int Mp314_PlayRandom(uint16_t start_time);
int Mp314_PlaySkip();
int Mp314_AtPlaying(FIL *fp_dir, FIL *fp_snd, uint8_t *FileRwBuf, uint16_t dirno, uint8_t trackno, BOOLEAN random_flag);
int Mp314_CmdActive();
int Mp314_CmdVolUp();
int Mp314_CmdVolDown();
int Mp314_CmdNext();
int Mp314_CmdPlayStop();
int Mp314_CmdDataTimestamp();
int Mp314_CmdShutdown();
int8 Mp314_RecieveTest();
void DispImage(uint16_t *image, int line, BOOLEAN inverse_flg);
void DispMsg(FIL *fp, uint8_t *buf, int line, char *msg, BOOLEAN inverse_flg);
int8 PauseAndSelectItem();
void debug_print(char *str);
void debug_print1(char *s, int16 val);
void debug_print1x(char *s, int16 val);
void debug_print2(char *s, int16 val, int16 val2);
int StrChr(char *s, char c);
FRESULT GetFontDataString(FIL *fp, uint8_t *buf, char *msg);
void GetFontData(FIL *Fil, uint8_t c, uint16_t *fontdata);
void TransAxisBytes(uint8_t *from, uint8_t *to);

#include "spi_sd.c"
#include "ff.c"

int main(int argc, char** argv) {
    //int         flg;
    FRESULT     res;
    FIL         Fil;
    FIL         Fil2;
    uint8_t     FileRwBuf[FILERWBUFSIZE];
    uint8_t     FileRwBuf2[FILERWBUFSIZE];
    char        ssid[20];
    char        passwd[20];
    char        mp3_14_ip[20];
    char        mp3_14_port[16];
    
    EVENT_SWDOWN = 0;
    EVENT_SWUP = 0;
    EVENT_SWSELECT = 0;
    EVENT_SWRETURN = 0;
    EVENT_SWPLAY = 0;
    EVENT_SWRANDOM = 0;
    EVENT_SWPOFF = 0;
    CHATTERING_TMR = 0;
    ConnectedMP3_14 = FALSE;
    
    InitApp();
    debug_print("debug_print\r\n");
    
    res = f_mount(&Fatfs, _T("0:"), 0);
    if (res == FR_OK){
        //OledTest();
        InitOled();
        
        if (ReadRouterFile(&Fil, FileRwBuf, FILERWBUFSIZE, ssid, passwd, mp3_14_ip, mp3_14_port) == 0){
            wifi_restart_point:
            if (wifi_connect_router(ssid, passwd) == 0){
                DispMsg(&Fil, FileRwBuf, 0, MSG_CONNECTED_ROUTER, FALSE);
                //debug_print("wifi_connect_router OK\r\n");
                for (int i=0; i<10; i++){
                    if (wifi_connect_server(mp3_14_ip, mp3_14_port) == WIFI_OK){
                        ConnectedMP3_14 = TRUE;
                        DispMsg(&Fil, FileRwBuf, 1, MSG_MP3_14_EXISTS, FALSE);
                        //debug_print("wifi_connect_server OK\r\n");
                        //Mp314_RecieveTest();
                        break;
                    } else {
                        DispMsg(&Fil, FileRwBuf, 1, MSG_MP3_14_NOT_EXISTS, FALSE);
                        debug_print("wifi_connect_server NG\r\n");
                        wait_ms(500);
                    }
                }
                if (ConnectedMP3_14 == TRUE){
                    GetFontDataString(&Fil, DigitImageData, DIGIT_CHAR);
                    if (GetMasterFiles(&Fil, FileRwBuf) == TRUE){
                        if (SelectDirectory(&Fil, &Fil2, FileRwBuf, FileRwBuf2) == WIFI_ERROR_DETECTED){
                            DispMsg(&Fil, FileRwBuf, 3, MSG_WIFI_ERROR, FALSE);
                            wait_ms(5000);
                            goto wifi_restart_point;
                        }
                    } else {
                        DispMsg(&Fil, FileRwBuf, 2, MSG_DATAFILE_IS_OLD, FALSE);
                        debug_print("Datafile is old\r\n");
                        wait_ms(5000);
                    }
                    SelectShutdown(&Fil, FileRwBuf);
                }
            } else {
                DispMsg(&Fil, FileRwBuf, 0, MSG_NO_ROUTER, FALSE);
                debug_print("wifi_connect_router NG\r\n");
                wait_ms(5000);
            }
            
        } else {
            DispMsg(&Fil, FileRwBuf, 0, MSG_SETTING_ERROR, FALSE);
            wait_ms(5000);
        }
    }
    
//    flg = 0;
//    for (int i=0; i<10; i++){
//        if (flg){
//            output_high(PIN_LED);
//        } else {
//            output_low(PIN_LED);
//        }
//        if (WifiTimeoutCounter > 1000){
//            flg = 1 - flg;
//            WifiTimeoutCounter = 0;
//        }
//    }
    
    output_low(PIN_POWER_ON_MAIN);
    while (1){
        
    }
    return 0;
}

void InitApp(){
    setup_oscillator(OSC_8MHZ | OSC_PLL_ON);
    // (1000us)/((4/32MHz)*250) = 32
    // -> 32*delay_cycle(250) = 1ms
    // (1000us)/((4/16MHz)*250) = 16
    // (1000us)/((4/8MHz)*250) = 8
    // (1000us)/((4/4MHz)*250) = 4
    WaitMsCycle = 16;
    
    /* Initialize peripherals */
    setup_adc(ADC_OFF);         // ADC disabled
    setup_ccp1(CCP_OFF);        // CCP1 disabled
    setup_ccp2(CCP_OFF);        // CCP2 disabled
    USBEN = 0;      // USB module disabled
    //UTRDIS = 1;     // On chip tranciever disabled
    
    output_high(PIN_POWER_ON_MAIN);
    output_float(PIN_DETECT_POWER_OFF);
    
    output_float(PIN_SWDOWN);
    output_float(PIN_SWUP);
    output_float(PIN_SWSELECT);
    output_float(PIN_SWRETURN);
    output_float(PIN_SWPLAY);
    output_float(PIN_SWRANDOM);
    port_b_pullups(0xCF);     // B0,1,2,3,6,7
    
    output_high(PIN_LED);
    delay_ms(500);      // チャッタリングをやり過ごす

    TRISA0 = 1;     // RP0=RX2
    TRISA1 = 0;     // RP1=TX2

    // Timer0 : (4/16MHz) * 1(prescaler) = 0.25us
    // 0.25us * 65536 = 16.3ms (TIMER0_INIT=0のとき)
    // 0.25us * (65536 - 15536) = 12.5ms (TIMER0_INIT=15536のとき)
    // 0.25us * (65536 - 55536) = 2.5ms
    // Timer0 : (4/8MHz) * 1(prescaler) = 0.50us
    // 0.50us * 65536 = 32.8ms (TIMER0_INIT=0のとき)
    // 0.50us * (65536 - 15536) = 25.0ms (TIMER0_INIT=15536のとき)
    // 0.50us * (65536 - 55536) = 5.0ms
    setup_timer_0(T0_INTERNAL | T0_DIV_16);
    set_timer0(TIMER0_INIT);
    clear_interrupt(INT_TIMER0);
    enable_interrupts(INT_TIMER0);
//    enable_interrupts(INT_RDA);       // RC1IEで設定する
    SendActiveCnt = 0;
    
#if 0
    //USART1の設定  ボーレート115200
    TXSTA1 = 0b00100100;    // 送信情報設定：非同期モード 8ビットデータ ノンパリティ //SYNC=0;BRGH=1;
    RCSTA1 = 0b10010000;    // 受信情報設定：Enable、８Bit、Continuous Receive Enable
    BAUDCON1 = 0b01001000;  //BRG16=1;    Receive operation is Idle、16-bit Baud Rate Generator – SPBRGHx and SPBRGx
    SPBRGH1 = 0;    // 16-bit Baud Rate Generator
    SPBRG1 = 34;    // 16-bit Baud Rate Generator
    RC1IF = 0;      // EUSART1 Receive Interrupt Flag bit
    RC1IE = 1;      // Enables the EUSART1 receive interrupt
#endif

    // USART2の設定 ボーレート115200
    // X = ((Fosc/Desired Baud Rate)/4) – 1
    //   = ((16000000/115200) / 4) - 1
    //   = 33.722
    TXSTA2 = 0b00100100;    // 送信情報設定：非同期モード 8ビットデータ ノンパリティ //SYNC=0;BRGH=1;
    RCSTA2 = 0b10010000;    // 受信情報設定：Enable、８Bit、Continuous Receive Enable
    BAUDCON2 = 0b01001000;  //BRG16=1;    Receive operation is Idle、16-bit Baud Rate Generator – SPBRGHx and SPBRGx
    SPBRGH2 = 0;    // 16-bit Baud Rate Generator
    SPBRG2 = 34;    // 16-bit Baud Rate Generator
    RC2IF = 0;      // EUSART2 Receive Interrupt Flag bit
    RC2IE = 1;      // Enables the EUSART2 receive interrupt
    
    PEIE = 1;       // Peripheral Interrupt Enable bit
    enable_interrupts(GLOBAL);
    
    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
}

void wait_ms(unsigned int16 ms){
    unsigned int16    i;
    unsigned int16    m;
    
    for (m=0; m<ms; m++){
        
        // 1ms
        for (i=0; i<WaitMsCycle; i++){
            delay_cycles(250);
        }
    }
}

// タイマーイベントハンドラ
#INT_TIMER0
void isr_timer0(){
    uint16_t n;
    
    n = Timer1;
    if (n) Timer1 = --n;
    n = Timer2;
    if (n) Timer2 = --n;
    
    if (CHATTERING_TMR > 0){
        CHATTERING_TMR++;
        if (CHATTERING_TMR == CHATTERING_ADJUST) CHATTERING_TMR = 0;
    } else {
        if (sw_down() == 0 && EVENT_SWDOWN == 0){
            EVENT_SWDOWN = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_up() == 0 && EVENT_SWUP == 0){
            EVENT_SWUP = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_select() == 0 && EVENT_SWRETURN == 0){
            EVENT_SWSELECT = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_return() == 0 && EVENT_SWSELECT == 0){
            EVENT_SWRETURN = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_play() == 0 && EVENT_SWPLAY == 0){
            EVENT_SWPLAY = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_random() == 0 && EVENT_SWRANDOM == 0){
            EVENT_SWRANDOM = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
        if (sw_poff() == 0 && EVENT_SWPOFF == 0){
            EVENT_SWPOFF = 1;
            CHATTERING_TMR = 1;     // chattering timer start
        }
    }
    
    WifiTimeoutCounter++;
    SendActiveCnt++;
    set_timer0(TIMER0_INIT);
}


// ルータに接続
int wifi_connect_router(char *ssid, char *passwd){
    char  buf[80];
    int     ret;
    
    wifi_readclear();
    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
    
    // echo off
    wifi_putstr("ATE0\r\n");
    ret = wifi_waitOk(500);
    debug_print1("ATE0 ", ret);
    if (ret != WIFI_OK){
        debug_print("ATE0 NG\r\n");
        return -1;
    } else {
        //debug_print("ATE0 OK\r\n");
    }
    
#if 0       // 不要
    // Tests AT Startup
    wifi_putstr("AT\r\n");
    ret = wifi_waitOk(500);
    if (ret != WIFI_OK){
        debug_print("AT NG\r\n");
        return -1;
    } else {
        //debug_print("AT OK\r\n");
    }
#endif
    
    // Wi-Fi Mode = Station
    wifi_putstr("AT+CWMODE=1\r\n");
    ret = wifi_waitOk(500);
    if (ret != WIFI_OK){
        debug_print("AT+CWMODE=1 NG\r\n");
        return -1;
    } else {
        //debug_print("AT+CWMODE=1 OK\r\n");
    }

    // Connects to an AP
    sprintf(buf, "AT+CWJAP=\"%s\",\"%s\"\r\n", ssid, passwd);
    wifi_putstr(buf);
    ret = wifi_waitOk(500);
    debug_print1("AT+CWJAP ", ret);
    if (ret != WIFI_OK){
        debug_print("AT+CWJAP NG\r\n");
        return -1;
    } else {
        //debug_print("AT+CWJAP OK\r\n");
    }
    
#if 0
    wifi_putstr("AT+CIFSR\r\n");        // Gets the Local IP Address
    if (wifi_waitOk(1000) == -1){
        DebugPuts("TIMEOUT \r\n");
    }
#endif
    
    return 0;
}

// サーバーに接続
int wifi_connect_server(const char *remote_ip, char *remote_port){
    char buf[80];
    int     ret;
    
    sprintf(buf, "AT+CIPSTART=\"TCP\",\"%s\",%s\r\n", remote_ip, remote_port);
    debug_print(buf);
    wifi_putstr(buf);
    ret = wifi_waitOk(100);
    debug_print1("AT+CIPSTART ", ret);
    if (ret != WIFI_OK){
        debug_print("AT+CIPSTART NG\r\n");
        return ret;
    }
    //debug_print("AT+CIPSTART OK\r\n");
    
    return WIFI_OK;
}

// サーバーにコマンド送信
int wifi_send_command(char* cmd, int bytes){
    char buf[80];
    int     ret;

    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
    PromptFlag = 0;
    sprintf(buf, "AT+CIPSEND=%d\r\n", bytes);
    //debug_print(buf);
    wifi_putstr(buf);
    ret = wifi_waitPrompt(100);
    if (ret != WIFI_OK){
        debug_print1("AT+CIPSEND NG2", ret);
        return -1;
    }
    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
    RecvedFlag = 0;
    for(int i = 0; i < bytes; i++){
        usartCharSend(cmd[i]);
    }
    ret = wifi_waitOk(100);
    if (ret != WIFI_OK){
        debug_print("AT+CIPSEND NG3\r\n");
        return -1;
    }
//    debug_print("AT+CIPSEND OK\r\n");
    
    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
    return WIFI_OK;
}

// Reset
int wifi_restart(){
    int     ret;
    
    wifi_putstr("AT+RST\r\n");
    ret = wifi_waitOk(100);
    debug_print1("AT+RST ", ret);
    if (ret != WIFI_OK){
        return ret;
    }
    //debug_print("AT+RST OK\r\n");
    
    return WIFI_OK;
}

//USART 文字列送信
void wifi_putstr(char *str){
    int     i;
    
    WifiSrIdx = 0;
    IPDStatus = IPD_INIT;
    RecvedFlag = 0;
    for (i=0; str[i] != '\0'; i++){
        usartCharSend(str[i]);
    }
}

int wifi_waitOk(long timeout){
    WifiTimeoutCounter = 0;
    while (RecvedFlag == 0 && WifiTimeoutCounter < timeout){
    }
    if (WifiTimeoutCounter >= timeout){
        return WIFI_TIMEOUT;
    } else if (RecvedFlag == 2){
        return WIFI_ERROR;
    } else if (RecvedFlag == 3){
        return WIFI_FULL;
    }
    
    return WIFI_OK;
}

int wifi_waitPrompt(long timeout){
    WifiTimeoutCounter = 0;
    while (PromptFlag == 0 && WifiTimeoutCounter < timeout){
        
    }
    if (WifiTimeoutCounter >= timeout){
        return WIFI_TIMEOUT;
    }
    return WIFI_OK;
}

void wifi_readclear(){
    int     c;
    
    while (RC2IF == 1){
        c = RCREG2;
    }
}


#int_rda2
void serial_rda_interrupt(){
    int c;
    
    if (RC2IF == 1){
        c = RCREG2;
        if(WifiSrIdx < (WIFISRBUFSIZE - 1)){
            WifiSrBuf[WifiSrIdx++] = c;
            if (c == 0x0a && WifiSrBuf[WifiSrIdx - 2] == 0x0d){
                if (WifiSrIdx >= 4 && strncmp(&WifiSrBuf[WifiSrIdx - 4], "OK", 2) == 0){
                    RecvedBytes = WifiSrIdx;
                    RecvedFlag = 1;
                } else if (WifiSrIdx >= 6 && strncmp(&WifiSrBuf[WifiSrIdx - 6], "ERROR", 4) == 0){
                    RecvedBytes = WifiSrIdx;
                    RecvedFlag = 2;
                }
            } else if (c == '>'){
                PromptFlag = 1;
            } else if (IPDStatus == IPD_INIT && c == '+'){
                IPDStatus = IPD_RECVD_PLUS;
                WifiSrIdx = 0;
                WifiSrBuf[WifiSrIdx++] = c;
                WifiSrIdx_IPD = 0;      // + の位置
                WifiTimeoutCounter = 0;
                DatatimestampRecv = FALSE;
            } else if (IPDStatus == IPD_RECVD_PLUS && (WifiSrIdx - WifiSrIdx_IPD) == 5){
                if (strncmp(&WifiSrBuf[WifiSrIdx_IPD], "+IPD,", 5) == 0){
                    IPDStatus = IPD_RECVD_IPD;
                    WifiSrIdx_IPD = WifiSrIdx;     // , の次の位置
                } else {
                    IPDStatus = IPD_INIT;
                }
            } else if (IPDStatus == IPD_RECVD_IPD && c == ':'){
                IPDStatus = IPD_RECVD_LEN;
                WifiSrBuf[WifiSrIdx - 1] = '\0';    // : の位置
                IPDDataLen = atoi(&WifiSrBuf[WifiSrIdx_IPD]);
                WifiSrIdx_IPD = WifiSrIdx;      // : の次 ＝　データの開始位置
            } else if (IPDStatus == IPD_RECVD_LEN){ 
                if ((WifiSrIdx - WifiSrIdx_IPD) == IPDDataLen){
                    IPDStatus = IPD_RECVD_DATA;
                } else if (c == MP3_14_CMD_DATATIMESTAMP){
                    DatatimestampRecv = TRUE;
                }
            }
            
        } else {
            RecvedBytes = WifiSrIdx;
            RecvedFlag = 3;
        }
        
    }
//    output_low(PIN_LED);
}


//USART 1文字送信
void usartCharSend(char c){
    while(TX2IF == 0) ;//送信可能まで待つ
    TXREG2 = c;
}

void usart2CharSend(char c){
    while(TX2IF == 0) ;//送信可能まで待つ
    TXREG2 = c;
}

void debug_print(char *str){
    fprintf(DEBUG, str);
}

void debug_print1(char *s, int16 val){
    char    msg[32];
    sprintf(msg,"%s (%Ld)\r\n", s, val);
    debug_print(msg);
}

void debug_print1x(char *s, int16 val){
    char    msg[32];
    sprintf(msg,"%s (%Lx)\r\n", s, val);
    debug_print(msg);
}

void debug_print2(char *s, int16 val, int16 val2){
    char    msg[32];
    sprintf(msg,"%s (%Ld,%Ld)\r\n", s, val, val2);
    debug_print(msg);
}

void InitOled(){
    int8    ack;
    uint8_t    ret = 0;
    
    debug_print("InitOlde 0 \r\n");
    i2c_start(OLED);
    ack = i2c_write(OLED, OLED_ADDRESS << 1);
    if (ack == 1){
        ret = 1;
    debug_print("InitOlde 1 \r\n");
        goto stop_condition;
    }
    debug_print("InitOlde 2 \r\n");
    ack = i2c_write(OLED, OLED_CB_COMMAND);    
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    ack = i2c_write(OLED, 0x8D);    // Set charge pump
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    ack = i2c_write(OLED, 0x14);    // Enable charge pump
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    ack = i2c_write(OLED, 0xAF);    // Display ON
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    ack = i2c_write(OLED, 0x20);    // Set memory addressing mode    
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    ack = i2c_write(OLED, 0x02);    // Page addressing mode
    if (ack == 1){
        ret = 1;
        goto stop_condition;
    }
    
stop_condition:
    i2c_stop(OLED);
}

void OledTest(){
    int16     i;
    int     page;
    
    InitOled();
    page = 0;
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_DATA);
    for (i=0; i<128; i++){
        i2c_write(OLED, i);
    }
    i2c_stop(OLED);
    
    page++;
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0xB0 | page);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x00);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x10);
    i2c_stop(OLED);
    
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_DATA);
    for (i=0; i<128; i++){
        i2c_write(OLED, 0x55);
    }
    i2c_stop(OLED);
    
}

/*
 * ルーター設定ファイル取得
 * 1行目：SSID
 * 2行目：パスワード
 * 3行目：MP3_14 IPアドレス
 * 4行目：MP3_14 ポート番号
 * 改行コード=CR+LF
 */
int8 ReadRouterFile(FIL *Fil, uint8_t *buf, int16 bufsize, char *ssid, char *passwd, char *mp3_14_ip, char *mp3_14_port){
    FRESULT     res;
    UINT        br;
    char        *c;
    int         i;
    int8        ret = 0;
    
    res = f_open(Fil, ROUTERFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        memset(buf, 0, bufsize);
        res = f_read(Fil, buf, bufsize, &br);
        if (res == FR_OK){
            c = buf;
            i = StrChr(c, 0x0d);
            if (i > 0){
                memcpy(ssid, c, i);
                ssid[i] = '\0';
                c += (i + 2);
            } else {
                ret = 1;
            }

            if (ret == 0){
                i = StrChr(c, 0x0d);
                if (i > 0){
                    memcpy(passwd, c, i);
                    passwd[i] = '\0';
                    c += (i + 2);
                } else {
                    ret = 1;
                }
            }
            
            if (ret == 0){
                i = StrChr(c, 0x0d);
                if (i > 0){
                    memcpy(mp3_14_ip, c, i);
                    mp3_14_ip[i] = '\0';
                    c += (i + 2);
                } else {
                    ret = 1;
                }
            }
            
            if (ret == 0){
                i = StrChr(c, 0x0d);
                if (i > 0){
                    memcpy(mp3_14_port, c, i);
                    mp3_14_port[i] = '\0';
                    c += (i + 2);
                } else {
                    ret = 1;
                }
            }
            
        }
    }
    

    f_close(Fil);
    
    return ret;
}

/*
 *  マスタファイルのタイムスタンプ比較。異なるときFALSEを返す。
 */
BOOLEAN GetMasterFiles(FIL *Fil, uint8_t *FileRwBuf){
//    BOOLEAN getflag = FALSE;
    char    ts_host[TIMESTAMP_LEN + 1];
    char    ts_local[TIMESTAMP_LEN + 1];
    FRESULT     res;
    UINT        br;
    int         status;
    int         timeoutcnt;
    
    // 再生中でもタイムスタンプ取得可能
//    while (1){
//        if (IPDStatus == IPD_RECVD_DATA){
//            if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_WAITING){
//                getflag = TRUE;
//                break;
//            } else if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_NOWPLAYING){
//                break;
//            }
//        }
//    }
//    
//    if (getflag == FALSE) return;
    
    // ホスト側データファイルのタイムスタンプ取得
//    debug_print("GetMasterFiles-----------------------------------\r\n");
//    wait_ms(5000);    // TeraTermの端末リセットのための時間
    WifiTimeoutCounter = 0;
    status = 0;
    while (1){
        if (IPDStatus == IPD_RECVD_DATA){
            debug_print("IPD_RECVD_DATA\r\n");
            if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_DATATIMESTAMP){
                memcpy(ts_host, &WifiSrBuf[WifiSrIdx_IPD + 1], TIMESTAMP_LEN);
                ts_host[TIMESTAMP_LEN] = '\0';
                break;
            } else {
                debug_print1x(" val=", WifiSrBuf[WifiSrIdx_IPD]);
                IPDStatus = IPD_INIT;
                WifiSrIdx = 0;
                if (status == 0){
                    if (Mp314_CmdDataTimestamp() == WIFI_OK){
                        debug_print("1\r\n");
                        status = 1;
                        timeoutcnt = 0;
                    } else {
                        debug_print("2\r\n");
                    }
                } else {    // ホストからの応答なし
                    timeoutcnt++;
                    if (timeoutcnt > 3){
                        status = 0;
                    }
                }
            }
//        } else if (IPDStatus != IPD_INIT){
//            debug_print1("3", IPDStatus);
//            debug_print1x("4", WifiSrBuf[WifiSrIdx_IPD]);
//        } else {
//            debug_print("wait ts_host\r\n");
//            wait_ms(500);
        }
        if (DatatimestampRecv == TRUE){
            debug_print(" RECVED \r\n");
            DatatimestampRecv = FALSE;
        }
        if (WifiTimeoutCounter > 2000){     // timeout : 10s
            break;
        }
    }
//    wait_ms(5000);
    debug_print("ts_host\r\n");
    debug_print(ts_host);
    debug_print("end \r\n");
    
    // ローカル側データファイルのタイムスタンプ
    do {
        res = f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
        if (res != FR_OK){
            char    msg[16];
            sprintf(msg,"f_open %d\r\n", res);
            debug_print(msg);
        }
        wait_ms(100);
    } while (res != FR_OK);
    debug_print("f_open DIRLISTFILE \r\n");
//    res = f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(Fil, FileRwBuf, 18, &br);
        if (res == FR_OK){
            memcpy(ts_local, &FileRwBuf[4], TIMESTAMP_LEN);
            ts_local[TIMESTAMP_LEN] = '\0';
        }
        f_close(Fil);
    } else {
        debug_print1("f_open ", res);
    }
    
    debug_print("ts_local\r\n");
    debug_print(ts_local);
    debug_print("end\r\n");
    
    // タイムスタンプ比較
    //if (strncmp(ts_host, ts_local, TIMESTAMP_LEN) == 0){
    if (memcmp(ts_host, ts_local, TIMESTAMP_LEN) == 0){
        return TRUE;
    }
    
    // データファイル取得の実装はしない。mp3_14側の改修も必要なので
    return FALSE;
}

/*
 *  ディレクトリ選択操作
 */
int8 SelectDirectory(FIL *Fil, FIL *Fil2, uint8_t *FileRwBuf, uint8_t *FileRwBuf2){
    FRESULT     res;
    FRESULT     res1;
    FRESULT     res2;
    DIRLIST     dir_parent;
    DIRLIST     dir_child;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;
    uint16_t    numSubDir;
    uint16_t    idxSubDir;          // ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurrentSubDirItem;   // 現在表示ページで選択中のDIR(0,1,2,3)
    int8        ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[NUM_OLED_LINES];
    uint16_t    numSubDirTab[NUM_OLED_LINES];
    uint16_t    numSndFile[NUM_OLED_LINES];
    uint16_t    recNoFirstSubdir[NUM_OLED_LINES];
    uint16_t    recNoFirstSndList[NUM_OLED_LINES];
    BOOLEAN     readdlflg;      // DIR LIST ファイルを読む必要があるか？
    uint8_t     i;
    UINT        br;

    ret = FR_OK;

    debug_print("SelectDirectory\r\n");
    do {
        res = f_open(Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
        if (res != FR_OK){
            char    msg[16];
            sprintf(msg,"f_open %d\r\n", res);
            debug_print(msg);
        }
        wait_ms(100);
    } while (res != FR_OK);
    debug_print("f_open DIRLISTFILE \r\n");
    
    // ヘッダレコードを飛ばす
    res = f_lseek(Fil, RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
        return res;
    }
    debug_print("f_lseek\r\n");
    
    // ルートDIR情報取得
    res = f_read_DIRLISTFILE(Fil, FileRwBuf, &dir_parent);
    if (res != FR_OK){
        return res;
    }
    debug_print("f_read_DIRLISTFILE\r\n");
    
    ofsSubDirStart = (uint32_t)dir_parent.recno_subdir_start * RECLEN_DIRLIST_FULL;        // サブDIRレコードの開始ファイル位置
    numSubDir = dir_parent.num_subdir;
    idxSubDir = 0;
    idxDirDepth = 0;
    idxCurrentSubDirItem = 0;
    
    // ルートDIRレコードの終端に移動（ヘッダとルートDIRレコードを飛ばす）
    res = f_lseek(Fil, 2*RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
        return res;
    }
    readdlflg = TRUE;
    
    // 音源ファイルリストもオープン
    do {
        res = f_open(Fil2, SNDFILELISTFILE, FA_READ | FA_OPEN_EXISTING);
        if (res != FR_OK){
            char    msg[16];
            sprintf(msg,"f_open %d\r\n", res);
            debug_print(msg);
        }
        wait_ms(100);
    } while (res != FR_OK);
    
    while (1){
        if (readdlflg == TRUE){
            debug_print1("read", dir_parent.num_subdir);
            for (i=0; i<NUM_OLED_LINES; i++){
                if ((idxSubDir + i) < dir_parent.num_subdir){
                    if ((res1 = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + i) * RECLEN_DIRLIST_FULL)) == FR_OK &&
                        (res2 = f_read_DIRLISTFILE(Fil, FileRwBuf, &dir_child)) == FR_OK){

                        dirno[i] = dir_parent.recno_subdir_start + idxSubDir + i;
                        numSubDirTab[i] = dir_child.num_subdir;
                        numSndFile[i] = dir_child.num_sndfile;
                        recNoFirstSubdir[i] = dir_child.recno_subdir_start;
                        recNoFirstSndList[i] = dir_child.recno_mp3list;
                        
                        if (i==idxCurrentSubDirItem){
                            res = f_read(Fil, FileRwBuf2, RECLEN_IMAGEDATA, &br);       // 反転表示する行だけ別のバッファに取得
                            DispImage((uint16_t *)FileRwBuf2, i, TRUE);
                        } else {
                            res = f_read(Fil, FileRwBuf, RECLEN_IMAGEDATA, &br);
                            DispImage((uint16_t *)FileRwBuf, i, FALSE);
                        }
                    } else {
                        //エラー
                        debug_print2("read", res1, res2);
                    }
                } else {
                    dirno[i] = 0;
                    numSubDirTab[i] = 0;
                    numSndFile[i] = 0;
                    recNoFirstSubdir[i] = 0;
                    recNoFirstSndList[i] = 0;
                    
                    memset(FileRwBuf, 0, RECLEN_IMAGEDATA);
                    DispImage((uint16_t *)FileRwBuf, i, FALSE);
                }
            }
            readdlflg = FALSE;
        }

        ret = PauseAndSelectItem();
        debug_print1x("PauseAndSelectItem", ret);
        if (ret == SWDOWN_PRESSED){
            debug_print2("SWDOWN_PRESSED", idxCurrentSubDirItem, numSubDir);
            idxCurrentSubDirItem++;
            if (idxCurrentSubDirItem == NUM_OLED_LINES && (idxSubDir + idxCurrentSubDirItem) < numSubDir){     // 次のページへ
                idxSubDir += NUM_OLED_LINES;
                idxCurrentSubDirItem = 0;
                readdlflg = TRUE;            
            } else if ((idxSubDir + idxCurrentSubDirItem) == numSubDir){    // 最後のページの最後の項目
                idxCurrentSubDirItem--;     // 進めない
                readdlflg = FALSE;            
            } else {    // 同一ページの次項目選択
                DispImage((uint16_t *)FileRwBuf2, idxCurrentSubDirItem - 1, FALSE);
                if (f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurrentSubDirItem) * RECLEN_DIRLIST_FULL) == FR_OK &&
                    f_read_DIRLISTFILE(Fil, FileRwBuf, &dir_child) == FR_OK &&
                    f_read(Fil, FileRwBuf2, RECLEN_IMAGEDATA, &br) == FR_OK){
                    DispImage((uint16_t *)FileRwBuf2, idxCurrentSubDirItem, TRUE);
                    readdlflg = FALSE;
                } else {
                    readdlflg = TRUE;
                }
            }
            debug_print2("SWDOWN_PRESSED 2", idxCurrentSubDirItem, readdlflg);
        } else if (ret == SWUP_PRESSED){
            if (idxCurrentSubDirItem == 0){
                if (idxSubDir >= NUM_OLED_LINES){      // 前ページへ
                    idxSubDir -= NUM_OLED_LINES;
                    idxCurrentSubDirItem = NUM_OLED_LINES - 1;
                    readdlflg = TRUE;
                } else {    // 進めない
                    readdlflg = FALSE;
                }
            } else {    // 同一ページの前項目選択
                idxCurrentSubDirItem--;
                DispImage((uint16_t *)FileRwBuf2, idxCurrentSubDirItem + 1, FALSE);
                if (f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurrentSubDirItem) * RECLEN_DIRLIST_FULL) == FR_OK &&
                    f_read_DIRLISTFILE(Fil, FileRwBuf2, &dir_child) == FR_OK &&
                    f_read(Fil, FileRwBuf2, RECLEN_IMAGEDATA, &br) == FR_OK){
                    DispImage((uint16_t *)FileRwBuf2, idxCurrentSubDirItem, TRUE);
                    readdlflg = FALSE;
                } else {
                    readdlflg = TRUE;
                }
                //readdlflg = TRUE;
            }
        } else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
            debug_print2("SWSELECT_PRESSED", idxCurrentSubDirItem, recNoFirstSubdir[idxCurrentSubDirItem]);
            if (recNoFirstSubdir[idxCurrentSubDirItem] > 1){
                // 選択中のITEM -> 親ITEMになる
                res = f_lseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurrentSubDirItem) * RECLEN_DIRLIST_FULL);
                res = f_read_DIRLISTFILE(Fil, FileRwBuf, &dir_parent);
                // 選択中のITEM
                rowSubDirStart = recNoFirstSubdir[idxCurrentSubDirItem];
                ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
                numSubDir = numSubDirTab[idxCurrentSubDirItem];     //*(RecDirList + DL_NUM_SUBDIR);
                // push current idxSubDir to stack
                stackIdxSubDir[idxDirDepth] = idxSubDir;
                stackIdxCurrentSubDirItem[idxDirDepth] = idxCurrentSubDirItem;
                idxDirDepth++;
                idxSubDir = 0;
                idxCurrentSubDirItem = 0;
                debug_print1("SWSELECT_PRESSED 2", ofsSubDirStart);
            } else if (numSndFile[idxCurrentSubDirItem] != 0){      // Have a MP3 file ?
                ret = SelectMp3InDirectory(Fil, Fil2, FileRwBuf, FileRwBuf2, recNoFirstSndList[idxCurrentSubDirItem], dirno[idxCurrentSubDirItem], numSndFile[idxCurrentSubDirItem], 0);
                if (ret == POWEROFF){
                    break;
                } else if (ret == WIFI_ERROR_DETECTED){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
            }
            readdlflg = TRUE;
        } else if (ret == SWRETURN_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
            uint16_t    r;

            // 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
            if (dir_child.recno_grandparentdir == dir_child.recno_parentdir){
                debug_print("SWRETURN_PRESSED top\r\n");
            } else {
                //r = *((uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR));   // rec no of grand parent
                r = dir_child.recno_grandparentdir;
                debug_print1("SWRETURN_PRESSED ", r);
                // Go to grand parent record
                res = f_lseek(Fil, (uint32_t)r * RECLEN_DIRLIST_FULL);
                if (res != FR_OK){
                    debug_print1("SWRET_PR lseek", res);
                    return res;
                }
                // Read grand parent record
                res = f_read_DIRLISTFILE(Fil, FileRwBuf, &dir_parent);
                if (res != FR_OK) return res;
                // the first "uncle" record
                //r = *((uint16_t *)(RecDirList + DL_RECNO_SUBDIR_START));
                r = dir_parent.recno_subdir_start;
                if (r > 1){
                    rowSubDirStart = r;
                    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
                    numSubDir = dir_parent.num_subdir;
                    // pop idxSubDir from stack -> parent
                    idxDirDepth--;
                    idxSubDir = stackIdxSubDir[idxDirDepth];
                    idxCurrentSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
                    if (idxSubDir >= numSubDir) idxSubDir = 0;
                } else {
                }
            }
            readdlflg = TRUE;
        } else if (ret == SWPLAY_PRESSED){      // PLAY
            if (numSndFile[idxCurrentSubDirItem] != 0){      // Have a MP3 file ?
                if (Mp314_PlayDir(dirno[idxCurrentSubDirItem], 0, 0) == WIFI_OK){
                    ret = Mp314_AtPlaying(Fil, Fil2, FileRwBuf, dirno[idxCurrentSubDirItem], 0, FALSE);
                    if (ret == SWPOFF_PRESSED){
                        ret = POWEROFF;
                        break;
                    } else if (ret == WIFI_ERROR_DETECTED){
                        ret = WIFI_ERROR_DETECTED;
                        break;
                    }
                } else {
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                readdlflg = TRUE;
            }
        } else if (ret == SWRANDOM_PRESSED){
            if (Mp314_PlayRandom(0) == WIFI_OK){
                ret = Mp314_AtPlaying(Fil, Fil2, FileRwBuf, dirno[idxCurrentSubDirItem], 0, TRUE);
                if (ret == SWPOFF_PRESSED){
                    ret = POWEROFF;
                    break;
                } else if (ret == WIFI_ERROR_DETECTED){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
            } else {
                ret = WIFI_ERROR_DETECTED;
                break;
            }
            readdlflg = TRUE;
        } else if (ret == SWPOFF_PRESSED){
            ret = POWEROFF;
            break;
        } else if (ret == NOWPLAYING_DETECTED){
            uint16_t    dirno_tmp = 0;
            uint8_t     trackno_tmp = 0;
            if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_NOWPLAYING){
                dirno_tmp = WifiSrBuf[WifiSrIdx_IPD + 1] | ((uint16_t)WifiSrBuf[WifiSrIdx_IPD + 2] << 8);
                trackno_tmp = WifiSrBuf[WifiSrIdx_IPD + 3];
            }
            ret = Mp314_AtPlaying(Fil, Fil2, FileRwBuf, dirno_tmp, trackno_tmp, FALSE);
            if (ret == SWPOFF_PRESSED){
                ret = POWEROFF;
                break;
            } else if (ret == WIFI_ERROR_DETECTED){
                ret = WIFI_ERROR_DETECTED;
                break;
            }
            readdlflg = TRUE;
        } else if (ret == WIFI_ERROR_DETECTED){
            ret = WIFI_ERROR_DETECTED;
            break;
        }
    }
    
    f_close(Fil);
    f_close(Fil2);
    return ret;
}

/*
 * ディレクトリ構成ファイルのカレントディレクトリ
 */
FRESULT f_read_DIRLISTFILE (FIL* fp, void* buff, DIRLIST* diritem){
    FRESULT res1;
    UINT    br;
    int     cnt;
    DWORD   pos;
    
    cnt = 0;
    pos = f_tell(fp) + DL_RECNO_SUBDIR_START;
    do {
        res1 = f_lseek(fp, pos);
        debug_print1(" f_read_DIR f_lseek", res1);
        if (res1 != FR_OK){
            res1 = f_lseek(fp, 0);
            if (res1 != FR_OK) break;
            wait_ms(1000);
        }
    } while (res1 != FR_OK);
    if (res1 != FR_OK) return FR_NO_FILE;
    res1 = f_read(fp, buff, RECLEN_DIRLIST - DL_RECNO_SUBDIR_START, &br);
    debug_print1(" f_read_DIR f_read", res1);
    debug_print2(" f_read_DIR f_read", br, RECLEN_DIRLIST - DL_RECNO_SUBDIR_START);
    if (br != (RECLEN_DIRLIST - DL_RECNO_SUBDIR_START)) return FR_NO_FILE;
    
    diritem->recno_subdir_start = *((uint16_t *)(buff));
    diritem->num_subdir = *(buff + DL_NUM_SUBDIR - DL_RECNO_SUBDIR_START);
    diritem->recno_parentdir = *((uint16_t *)(buff + DL_RECNO_PARENTDIR - DL_RECNO_SUBDIR_START));
    diritem->recno_grandparentdir = *((uint16_t *)(buff + DL_RECNO_GRANDPARENTDIR - DL_RECNO_SUBDIR_START));
    diritem->num_sndfile = *(buff + DL_NUM_SNDFILE - DL_RECNO_SUBDIR_START);
    diritem->recno_mp3list = *((uint16_t *)(buff + DL_RECNO_MP3LIST - DL_RECNO_SUBDIR_START));
    
    debug_print1(" f_read_DIR OK", res1);
    // fpはイメージデータの先頭に位置する
    return res1;
}

// 現在のサブディレクトリからMP3を選択する
int8 SelectMp3InDirectory(FIL *Fil, FIL *Fil2, uint8_t *FileRwBuf, uint8_t *FileRwBuf2, uint16_t recno_first, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno){
    FRESULT     res;
    int8        ret;
    uint8_t     trackno;        // 再生中の曲番 (0,1,2,...,numsndfile-1)
    uint8_t     idxtrack;       // 選択操作中のページ先頭の曲番 (0,1,2,...,numsndfile-1)
    uint16_t    idxCurrentTitleItem;
    BOOLEAN     readdlflg;      // MP3 LIST ファイルを読む必要があるか？
    UINT        br;
    int         i;

    ret = 0;
    
    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
    idxtrack = 0;
    idxCurrentTitleItem = 0;
    readdlflg = TRUE;
    
    while (1){
        if (readdlflg == TRUE){
            for (i=0; i<NUM_OLED_LINES; i++){
                if ((idxtrack + i) < numsndfile){
                    f_lseek(Fil2, (uint32_t)(recno_first + idxtrack + i)*(RECLEN_MP3LIST_FULL) + RECLEN_MP3LIST);
                    if (i==idxCurrentTitleItem){
                        res = f_read(Fil2, FileRwBuf2, RECLEN_IMAGEDATA, &br);      // 反転表示行だけ別のバッファに取得
                        DispImage((uint16_t *)FileRwBuf2, i, TRUE);
                    } else {
                        res = f_read(Fil2, FileRwBuf, RECLEN_IMAGEDATA, &br);
                        DispImage((uint16_t *)FileRwBuf, i, FALSE);
                    }
                } else {
                    memset(FileRwBuf, 0, RECLEN_IMAGEDATA);
                    DispImage((uint16_t *)FileRwBuf, i, FALSE);
                }
            }
            readdlflg = FALSE;
        }

        loop_selectmp3:
        ret = PauseAndSelectItem();
        if (ret == SWDOWN_PRESSED){         // 同一ページの次項目選択
            idxCurrentTitleItem++;
            if (idxCurrentTitleItem == NUM_OLED_LINES && (idxtrack + idxCurrentTitleItem) < numsndfile){
                idxtrack += NUM_OLED_LINES;
                idxCurrentTitleItem = 0;
                readdlflg = TRUE;            
            } else if ((idxtrack + idxCurrentTitleItem) == numsndfile){
                idxCurrentTitleItem++;
                readdlflg = FALSE;
            } else {                    // 同一ページの次項目選択
                DispImage((uint16_t *)FileRwBuf2, idxCurrentTitleItem - 1, FALSE);
                if (f_lseek(Fil2, (uint32_t)(recno_first + idxtrack + idxCurrentTitleItem)*(RECLEN_MP3LIST_FULL) + RECLEN_MP3LIST) == FR_OK && 
                    f_read(Fil2, FileRwBuf2, RECLEN_IMAGEDATA, &br) == FR_OK){
                    DispImage((uint16_t *)FileRwBuf2, idxCurrentTitleItem, TRUE);
                    readdlflg = FALSE;            
                } else {
                    readdlflg = TRUE;            
                }
            }
            continue;
        } else if (ret == SWUP_PRESSED){  // 同一ページの前項目選択
            if (idxCurrentTitleItem == 0){
                if (idxtrack >=NUM_OLED_LINES){
                    idxtrack -= NUM_OLED_LINES;
                    idxCurrentTitleItem = NUM_OLED_LINES - 1;
                    readdlflg = TRUE;
                } else {
                    readdlflg = FALSE;
                }
            } else {        // 同一ページの前項目選択
                idxCurrentTitleItem--;
                DispImage((uint16_t *)FileRwBuf2, idxCurrentTitleItem + 1, FALSE);
                if (f_lseek(Fil2, (uint32_t)(recno_first + idxtrack + idxCurrentTitleItem)*(RECLEN_MP3LIST_FULL) + RECLEN_MP3LIST) == FR_OK && 
                    f_read(Fil2, FileRwBuf2, RECLEN_IMAGEDATA, &br) == FR_OK){
                    DispImage((uint16_t *)FileRwBuf2, idxCurrentTitleItem, TRUE);
                    readdlflg = FALSE;            
                } else {
                    readdlflg = TRUE;            
                }
            }
            continue;
        } else if (ret == SWPLAY_PRESSED){ 
            trackno = idxtrack + idxCurrentTitleItem;
            if (Mp314_PlayDir(dirno, trackno, 0) == WIFI_OK){
                ret = Mp314_AtPlaying(Fil, Fil2, FileRwBuf, dirno, 0, FALSE);
                if (ret == SWPOFF_PRESSED){
                    ret = POWEROFF;
                    break;
                } else if (ret == WIFI_ERROR_DETECTED){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                } else if (ret == 0){       // 正常終了
                    ret = PLAY_OTHER_DIR;
                    break;
                }
            } else {
                // TODO:
            }
            readdlflg = TRUE;
            continue;
        } else if (ret == SWRANDOM_PRESSED){
            if (Mp314_PlayRandom(0) == WIFI_OK){
                ret = Mp314_AtPlaying(Fil, Fil2, FileRwBuf, dirno, 0, TRUE);
                if (ret == SWPOFF_PRESSED){
                    ret = POWEROFF;
                    break;
                } else if (ret == WIFI_ERROR_DETECTED){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
            } else {
                ret = WIFI_ERROR_DETECTED;
                break;
            }
            readdlflg = TRUE;
            continue;
        } else if (ret == SWRETURN_PRESSED){
            ret = PLAY_OTHER_DIR;
            break;
        } else if (ret == SWPOFF_PRESSED){
            ret = POWEROFF;
            break;
        } else if (ret == WIFI_ERROR_DETECTED){
            ret = WIFI_ERROR_DETECTED;
            break;
        } else {
            continue;
        }
    }
    return (ret);
}

/*
 *  停止時にホストも遮断するかを選択
 */
void SelectShutdown(FIL *Fil, uint8_t *FileRwBuf){
    int         status = 0;
    int8        ret;
    
    DispMsg(Fil, FileRwBuf, 2, "                ", FALSE);
    DispMsg(Fil, FileRwBuf, 3, "                ", FALSE);
    while (1){
        DispMsg(Fil, FileRwBuf, 0, "HOST+HC OFF     ", (status==0 ? TRUE : FALSE));
        DispMsg(Fil, FileRwBuf, 1, "HC OFF only     ", (status==1 ? TRUE : FALSE));
        
        ret = PauseAndSelectItem();
        if (ret == SWDOWN_PRESSED){
            status = 1 - status;
        } else if (ret == SWUP_PRESSED){
            status = 1 - status;
        } else if (ret == SWSELECT_PRESSED){
            if (status == 0){
                Mp314_CmdShutdown();
            }
            break;
        }
    }
}


/*
 *  指定ディレクトリの特定曲番号から再生開始(MP3_14_CMD_PLAYDIR)
 */
void Mp314_PlayDir_OLD(uint16_t dirno, uint8_t trackno, uint16_t start_time){
    char   cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_PLAYDIR;
    cmd[1] = (uint8_t)(dirno & 0x00FF);
    cmd[2] = (uint8_t)((dirno >> 8) & 0x00FF);
    cmd[3] = trackno;
    cmd[4] = (uint8_t)(start_time & 0x00FF);
    cmd[5] = (uint8_t)((start_time >> 8) & 0x00FF);
    wifi_send_command(cmd, MP3_14_CMDLENGTH);
    
}

int8 Mp314_RecieveTest(){
    int8        ret;
    int8        i;
    
    ret = 0;
    debug_print("Mp314_RecieveTest\r\n");
    
//    wifi_putstr("AT+CIPDINFO=0\r\n");
//    ret = wifi_waitOk(100);
//    debug_print("OK\r\n");

    Mp314_CmdActive();
    
    while (1){
        if (IPDStatus == IPD_RECVD_DATA){
            debug_print1("IPD", IPDDataLen);
            
            for (i=0; i<IPDDataLen; i++){
                debug_print1x(" +", WifiSrBuf[i + WifiSrIdx_IPD]);
            }
            IPDStatus = IPD_INIT;
            //Mp314_CmdActive();
        } else if (IPDStatus != IPD_INIT){
            debug_print1(" IPDStatus", IPDStatus);
            debug_print2("          ", WifiSrIdx_IPD, WifiSrIdx);
        }
        if (SendActiveCnt > 1000 && IPDStatus == IPD_INIT){
            Mp314_CmdActive();
            SendActiveCnt = 0;
        }
        if (EVENT_SWPOFF ){
            EVENT_SWPOFF = 0;
            ret = SWPOFF_PRESSED;
            break;
        }
    }
    
    return ret;    
}

/*
 *  指定ディレクトリの特定曲番号から再生開始(MP3_14_CMD_PLAYDIR)
 */
int Mp314_PlayDir(uint16_t dirno, uint8_t trackno, uint16_t start_time){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_PLAYDIR;
    cmd[1] = (uint8_t)(dirno & 0x00FF);
    cmd[2] = (uint8_t)((dirno >> 8) & 0x00FF);
    cmd[3] = trackno;
    cmd[4] = (uint8_t)(start_time & 0x00FF);
    cmd[5] = (uint8_t)((start_time >> 8) & 0x00FF);
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}


int Mp314_PlayRandom(uint16_t start_time){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_PLAYRANDOM;
    cmd[1] = (uint8_t)(start_time & 0x00FF);
    cmd[2] = (uint8_t)((start_time >> 8) & 0x00FF);
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

// ランダム再生時に次のディレクトリの再生へ進む
int Mp314_PlaySkip(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_PLAYSKIP;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

/*
 *  再生中処理
 */
int Mp314_AtPlaying(FIL *fp_dir, FIL *fp_snd, uint8_t *FileRwBuf, uint16_t dirno, uint8_t trackno, BOOLEAN random_flag){
    FRESULT     res;
    DIRLIST     dir_child;
    UINT        br;
    int8        ret;
    int16       i, j;
    uint16_t    dirno_bk = 0;
    uint8_t     trackno_bk = 0;
    uint16_t    playing_time = 0;
    uint16_t    playing_time_sec = 0;
    uint16_t    playing_time_min = 0;
    uint16_t    total_time = 0;
    uint16_t    t1;
    uint8_t     num_sndfile = 0;
    char        str[32];
    int         len;
    
    if (dirno > 0){
        // Directory Name
        res = f_lseek(fp_dir, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
        res = f_read_DIRLISTFILE(fp_dir, FileRwBuf, &dir_child);
        res = f_read(fp_dir, FileRwBuf, RECLEN_IMAGEDATA, &br);
        DispImage((uint16_t *)FileRwBuf, 0, FALSE);
        num_sndfile = dir_child.num_sndfile;
        // Track Name
        res = f_lseek(fp_snd, (uint32_t)(dir_child.recno_mp3list + trackno)*(RECLEN_MP3LIST_FULL) + RECLEN_MP3LIST);
        res = f_read(fp_snd, FileRwBuf, RECLEN_IMAGEDATA, &br);
        DispImage((uint16_t *)FileRwBuf, 1, FALSE);
        memset(FileRwBuf, 0, RECLEN_IMAGEDATA);
        DispImage((uint16_t *)FileRwBuf, 2, FALSE);
        DispImage((uint16_t *)FileRwBuf, 3, FALSE);
    }
    
    while (1){
        if (IPDStatus == IPD_RECVD_DATA){
            debug_print("IPD_RECVD_DATA\r\n");
//            debug_print1("IPD", IPDDataLen);
//            debug_print2("          ", WifiSrIdx_IPD, WifiSrIdx);
            
//            for (i=0; i<IPDDataLen; i++){
//                debug_print1x(" +", WifiSrBuf[WifiSrIdx_IPD + i]);
//            }
            IPDStatus = IPD_INIT;
            
            if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_NOWPLAYING){
                dirno = WifiSrBuf[WifiSrIdx_IPD + 1] | ((uint16_t)WifiSrBuf[WifiSrIdx_IPD + 2] << 8);
                trackno = WifiSrBuf[WifiSrIdx_IPD + 3];
                playing_time = WifiSrBuf[WifiSrIdx_IPD + 4] | ((uint16_t)WifiSrBuf[WifiSrIdx_IPD + 5] << 8);
                total_time = WifiSrBuf[WifiSrIdx_IPD + 11] | ((uint16_t)WifiSrBuf[WifiSrIdx_IPD + 12] << 8);
                
                if (dirno != dirno_bk || trackno != trackno_bk){
                    // Directory Name
                    res = f_lseek(fp_dir, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
                    res = f_read_DIRLISTFILE(fp_dir, FileRwBuf, &dir_child);
                    res = f_read(fp_dir, FileRwBuf, RECLEN_IMAGEDATA, &br);
                    DispImage((uint16_t *)FileRwBuf, 0, FALSE);
                    num_sndfile = dir_child.num_sndfile;
                    // Track Name
                    res = f_lseek(fp_snd, (uint32_t)(dir_child.recno_mp3list + trackno)*(RECLEN_MP3LIST_FULL) + RECLEN_MP3LIST);
                    res = f_read(fp_snd, FileRwBuf, RECLEN_IMAGEDATA, &br);
                    DispImage((uint16_t *)FileRwBuf, 1, FALSE);
                    memset(FileRwBuf, 0, RECLEN_IMAGEDATA);
                    DispImage((uint16_t *)FileRwBuf, 2, FALSE);
                    DispImage((uint16_t *)FileRwBuf, 3, FALSE);
                    dirno_bk = dirno;
                    trackno_bk = trackno;
                }
                // FileRwBufに３行目の表示イメージを作成
                playing_time_sec = (playing_time % 60);
                playing_time_min = (playing_time - playing_time_sec) / 60;
                sprintf(str, "%3d/%3d  %3Lu:%02Lu ", trackno + 1, num_sndfile, playing_time_min, playing_time_sec);
                //debug_print(str);
                //debug_print("\r\n");                
                len = strlen(str);
                if (len > 16) len = 16;
                for (j=0; j<len; j++){
                    if (str[j] >= '0' && str[j] <= '9'){
                        t1 = str[j] - '0';
                    } else if (str[j] == ':'){
                        t1 = 10;
                    } else if (str[j] == '/'){
                        t1 = 12;
                    } else {
                        t1 = 11;    // ' '
                    }
                    for (i=0; i<FONT_BYTES_CHAR; i++){
                        FileRwBuf[j * FONT_BYTES_CHAR + i] = DigitImageData[t1 * FONT_BYTES_CHAR + i];
                    }
                }
                DispImage((uint16_t *)FileRwBuf, 2, FALSE);
                
                // 再生進行バー
                if (total_time > 0){
                    j = ((uint32_t)playing_time * LCD_WIDTH) / total_time;
                } else {
                    j = 0;
                }
                for (i=0; i<LCD_WIDTH; i++){
                    if (i < j){
                        FileRwBuf[2*i] = 0xFF;
                        FileRwBuf[2*i + 1] = 0xFF;
                    } else {
                        FileRwBuf[2*i] = 0x00;
                        FileRwBuf[2*i + 1] = 0x00;
                    }
                }
                DispImage((uint16_t *)FileRwBuf, 3, FALSE);
                
                // MP3_14_CMD_NOWPLAYINGの受信に同期してコントローラが実行中であることをホストに伝える
                if (Mp314_CmdActive() != WIFI_OK){
                    debug_print("Mp314_CmdActive ERR \r\n");
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
            } else if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_WAITING){
                ret = 0;
                break;
            }
        } else if (IPDStatus != IPD_INIT){
//            debug_print2(" IPDStatus", IPDStatus, IPDDataLen);
//            debug_print2("          ", WifiSrIdx_IPD, WifiSrIdx);
            if (WifiTimeoutCounter > 10){   // データ受信のタイムアウト
                IPDStatus = IPD_INIT;
            }
        } else if (IPDStatus == IPD_INIT){
            if (EVENT_SWDOWN ){
                if (Mp314_CmdVolDown() != WIFI_OK){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                EVENT_SWDOWN = 0;
            } else if (EVENT_SWUP ){
                if (Mp314_CmdVolUp() != WIFI_OK){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                EVENT_SWUP = 0;
            } else if (EVENT_SWPLAY){
                if (Mp314_CmdNext() != WIFI_OK){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                EVENT_SWPLAY = 0;
            } else if (EVENT_SWRETURN){
                if (Mp314_CmdPlayStop() != WIFI_OK){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                EVENT_SWRETURN = 0;
                ret = SWRETURN_PRESSED;
                break;
            } else if (EVENT_SWRANDOM && random_flag){
                if (Mp314_PlaySkip() != WIFI_OK){
                    ret = WIFI_ERROR_DETECTED;
                    break;
                }
                EVENT_SWRANDOM = 0;
            }
        }
        if (EVENT_SWPOFF ){
            EVENT_SWPOFF = 0;
            ret = SWPOFF_PRESSED;
            break;
        }
    }
    
    return ret;
}

//
int Mp314_CmdActive(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_ACTIVE;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

int Mp314_CmdVolUp(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_VOLUP;
    cmd[1] = MP3_14_VOLUME_STEP;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

int Mp314_CmdVolDown(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_VOLDOWN;
    cmd[1] = MP3_14_VOLUME_STEP;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

int Mp314_CmdNext(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_NEXT;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

int Mp314_CmdPlayStop(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_PLAYSTOP;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

// ホストのデータファイルのタイムスタンプ取得
int Mp314_CmdDataTimestamp(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_DATATIMESTAMP;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

// ホスト電源遮断
int Mp314_CmdShutdown(){
    char        cmd[MP3_14_CMDLENGTH];
    
    memset(cmd, 0, MP3_14_CMDLENGTH);
    cmd[0] = MP3_14_CMD_SHUTDOWN;
    return wifi_send_command(cmd, MP3_14_CMDLENGTH);
}

/*
 * image        : イメージデータ
 * line         : 行(0,1,2,3)
 * inverse_flg  : 反転するときTRUE
 */
void DispImage(uint16_t *image, int line, BOOLEAN inverse_flg){
    int16     i;
    int     page;
    
    page = 2*line;
    
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0xB0 | page);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x00);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x10);
    i2c_stop(OLED);
    
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_DATA);
    if (inverse_flg == FALSE){
        for (i=0; i<LCD_WIDTH; i++){
            i2c_write(OLED, image[i] & 0xFF);
        }
    } else {
        for (i=0; i<LCD_WIDTH; i++){
            i2c_write(OLED, ~image[i] & 0xFF);
        }
    }
    i2c_stop(OLED);
    
    page++;
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0xB0 | page);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x00);
    i2c_write(OLED, OLED_CB_COMMAND_CO);
    i2c_write(OLED, 0x10);
    i2c_stop(OLED);
    
    i2c_start(OLED);    
    i2c_write(OLED, OLED_ADDRESS << 1);
    i2c_write(OLED, OLED_CB_DATA);
    if (inverse_flg == FALSE){
        for (i=0; i<LCD_WIDTH; i++){
            i2c_write(OLED, (image[i] >> 8) & 0xFF);
        }
    } else {
        for (i=0; i<LCD_WIDTH; i++){
            i2c_write(OLED, (~image[i] >> 8) & 0xFF);
        }
    }
    i2c_stop(OLED);
}

// 選択画面における入力待ち
int8 PauseAndSelectItem(){
    int8     ret;
//    int     i;

loop_PauseAndSelectItem:
    if (IPDStatus == IPD_RECVD_DATA){
//        debug_print1("IPD", IPDDataLen);
//        for (i=0; i<IPDDataLen; i++){
//            debug_print1x(" +", WifiSrBuf[WifiSrIdx_IPD + i]);
//        }
        IPDStatus = IPD_INIT;

        if (WifiSrBuf[WifiSrIdx_IPD] == MP3_14_CMD_NOWPLAYING){
            return NOWPLAYING_DETECTED;
        }
    }
    if (SendActiveCnt > 1000){
        // コントローラが実行中であることをホストに伝える
        if (Mp314_CmdActive() != WIFI_OK){
            return WIFI_ERROR_DETECTED;
        }
        SendActiveCnt = 0;
    }
    if (EVENT_SWDOWN ){
        EVENT_SWDOWN = 0;
        return SWDOWN_PRESSED;
    }
    if (EVENT_SWUP ){
        EVENT_SWUP = 0;
        return SWUP_PRESSED;
    }
    if (EVENT_SWSELECT ){
        EVENT_SWSELECT = 0;
        return SWSELECT_PRESSED;
    }
    if (EVENT_SWRETURN ){
        EVENT_SWRETURN = 0;
        return SWRETURN_PRESSED;
    }
    if (EVENT_SWPLAY ){
        EVENT_SWPLAY = 0;
        return SWPLAY_PRESSED;
    }
    if (EVENT_SWRANDOM){
        EVENT_SWRANDOM = 0;
        return SWRANDOM_PRESSED;
    }
    if (EVENT_SWPOFF ){
        EVENT_SWPOFF = 0;
        return SWPOFF_PRESSED;
    }

    goto loop_PauseAndSelectItem;

    return ret;
}

/*
 * 文字列sから文字cが最初に現れる位置
 */
int StrChr(char *s, char c){
    int     i;
    
    for (i=0; s[i] != '\0'; i++){
        if (s[i] == c){
            return i;
        }
    }
    return -1;
}

/*
 *  固定文字列表示。半角文字だけの最大１６文字
 */
void DispMsg(FIL *fp, uint8_t *buf, int line, char *msg, BOOLEAN inverse_flg){
    if (GetFontDataString(fp, buf, msg) == FR_OK){
        DispImage((uint16_t *)buf, line, inverse_flg);        
    }
}

/*
 *  文字列のFONTデータ取得
 */
FRESULT GetFontDataString(FIL *fp, uint8_t *buf, char *msg){
    FRESULT     res;
    int         i;
    int         bytes_per_char = 16;
    int         max_len = 16;
    
    res = f_open(fp, FONTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        for (i=0; i<max_len && msg[i] != '\0'; i++){
            GetFontData(fp, msg[i], (uint16_t *)(buf + bytes_per_char*i));
        }
        f_close(fp);
    }
    return res;
}

/*
 *  半角文字cのフォントデータ取得
 *  高さ１６ Ｘ 幅８　固定
 *  fp : 半角文字のFONTXフォントファイル
 */
void GetFontData(FIL *fp, uint8_t c, uint16_t *fontdata){
    FRESULT     res;
    UINT        br;
    uint8_t     buff[32];
	uint8_t     work1[8];
	uint8_t     to1[8];
    uint8_t     to2[8];
    int         k;
    int         bytes_per_char = FONT_BYTES_CHAR;
    int         fontheight = FONT_HEIGHT;
    int         fontwidth = FONT_WIDTH;
    
    res = f_lseek(fp, 17 + (DWORD)bytes_per_char * c);
    if (res == FR_OK){
        res = f_read(fp, buff, bytes_per_char, &br);
        if (res == FR_OK){
			// 上ブロック
			for (k=0; k<fontwidth; k++){
				work1[k] = buff[k];
			}
			TransAxisBytes(work1, to1);
            
            // 下ブロック
            for (k=0; k<fontwidth; k++){
                if (k < (fontheight - 8)){
                    work1[k] = buff[k + 8];
                } else {
                    work1[k] = 0;
                }
            }
            TransAxisBytes(work1, to2);
            
            for (k=0; k<8; k++){
                *fontdata++ = (to1[k] | (((uint16_t)to2[k]) << 8));
            }
        } else {
            for (k=0; k<8; k++){
                *fontdata++ = 0;
            }
        }
    } else {
        for (k=0; k<8; k++){
            *fontdata++ = 0;
        }
    }
}

/*
 *	長さ８のバイト配列のビット位置を縦横入れ替え
 */
void TransAxisBytes(uint8_t *from, uint8_t *to){
	int		i;

	for (i=0; i<8; i++) to[i] = 0;
	
	to[0] |= ((from[0] & 0x80) >> 7);
	to[1] |= ((from[0] & 0x40) >> 6);
	to[2] |= ((from[0] & 0x20) >> 5);
	to[3] |= ((from[0] & 0x10) >> 4);
	to[4] |= ((from[0] & 0x08) >> 3);
	to[5] |= ((from[0] & 0x04) >> 2);
	to[6] |= ((from[0] & 0x02) >> 1);
	to[7] |= ((from[0] & 0x01) >> 0);

	to[0] |= ((from[1] & 0x80) >> 6);
	to[1] |= ((from[1] & 0x40) >> 5);
	to[2] |= ((from[1] & 0x20) >> 4);
	to[3] |= ((from[1] & 0x10) >> 3);
	to[4] |= ((from[1] & 0x08) >> 2);
	to[5] |= ((from[1] & 0x04) >> 1);
	to[6] |= ((from[1] & 0x02) >> 0);
	to[7] |= ((from[1] & 0x01) << 1);

	to[0] |= ((from[2] & 0x80) >> 5);
	to[1] |= ((from[2] & 0x40) >> 4);
	to[2] |= ((from[2] & 0x20) >> 3);
	to[3] |= ((from[2] & 0x10) >> 2);
	to[4] |= ((from[2] & 0x08) >> 1);
	to[5] |= ((from[2] & 0x04) >> 0);
	to[6] |= ((from[2] & 0x02) << 1);
	to[7] |= ((from[2] & 0x01) << 2);

	to[0] |= ((from[3] & 0x80) >> 4);
	to[1] |= ((from[3] & 0x40) >> 3);
	to[2] |= ((from[3] & 0x20) >> 2);
	to[3] |= ((from[3] & 0x10) >> 1);
	to[4] |= ((from[3] & 0x08) >> 0);
	to[5] |= ((from[3] & 0x04) << 1);
	to[6] |= ((from[3] & 0x02) << 2);
	to[7] |= ((from[3] & 0x01) << 3);

	to[0] |= ((from[4] & 0x80) >> 3);
	to[1] |= ((from[4] & 0x40) >> 2);
	to[2] |= ((from[4] & 0x20) >> 1);
	to[3] |= ((from[4] & 0x10) >> 0);
	to[4] |= ((from[4] & 0x08) << 1);
	to[5] |= ((from[4] & 0x04) << 2);
	to[6] |= ((from[4] & 0x02) << 3);
	to[7] |= ((from[4] & 0x01) << 4);

	to[0] |= ((from[5] & 0x80) >> 2);
	to[1] |= ((from[5] & 0x40) >> 1);
	to[2] |= ((from[5] & 0x20) >> 0);
	to[3] |= ((from[5] & 0x10) << 1);
	to[4] |= ((from[5] & 0x08) << 2);
	to[5] |= ((from[5] & 0x04) << 3);
	to[6] |= ((from[5] & 0x02) << 4);
	to[7] |= ((from[5] & 0x01) << 5);

	to[0] |= ((from[6] & 0x80) >> 1);
	to[1] |= ((from[6] & 0x40) >> 0);
	to[2] |= ((from[6] & 0x20) << 1);
	to[3] |= ((from[6] & 0x10) << 2);
	to[4] |= ((from[6] & 0x08) << 3);
	to[5] |= ((from[6] & 0x04) << 4);
	to[6] |= ((from[6] & 0x02) << 5);
	to[7] |= ((from[6] & 0x01) << 6);

	to[0] |= ((from[7] & 0x80) >> 0);
	to[1] |= ((from[7] & 0x40) << 1);
	to[2] |= ((from[7] & 0x20) << 2);
	to[3] |= ((from[7] & 0x10) << 3);
	to[4] |= ((from[7] & 0x08) << 4);
	to[5] |= ((from[7] & 0x04) << 5);
	to[6] |= ((from[7] & 0x02) << 6);
	to[7] |= ((from[7] & 0x01) << 7);

}
